from .scores import ScoreMethod
from .params import SABaseParams
from .scores import compute_score
from .scores import ComputedScore
from .scores import apply_roi
from .scores import get_disk_mask_radius



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "15/02/2017"


class OctaveH5Editor(object):
    """Abstract class inheritate by classes which will edit an h5 file"""

    def __init__(self):
        """ """
        self.loadedStructures = None

    def loadReconsParams(self, structures):
        """
        Load h5 structure from the given h5 file

        :param structures: the structures loaded
        """
        self.loadedStructures = structures

    def getLoadedStructures(self):
        """
        :return: the dictionnary of the loaded structure
        """
        return self.loadedStructures

    def getStructs(self):
        """
        :return: the dictionnary of all the h5 structure"""
        raise NotImplementedError("OctaveH5Editor is an abstract class")



# # coding: utf-8
# # /*##########################################################################
# #
# # Copyright (c) 2017 European Synchrotron Radiation Facility
# #
# # Permission is hereby granted, free of charge, to any person obtaining a copy
# # of this software and associated documentation files (the "Software"), to deal
# # in the Software without restriction, including without limitation the rights
# # to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# # copies of the Software, and to permit persons to whom the Software is
# # furnished to do so, subject to the following conditions:
# #
# # The above copyright notice and this permission notice shall be included in
# # all copies or substantial portions of the Software.
# #
# # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# # THE SOFTWARE.
# #
# # ###########################################################################*/
#
# __authors__ = ["H. Payno"]
# __license__ = "MIT"
# __date__ = "31/07/2020"
#
#
# import unittest
# from tomwer.test.utils import UtilsTest
# import h5py
# import shutil
# import tempfile
# from tomwer.core.utils.scanutils import MockBlissAcquisition
# from tomwer.core.scan.blissscan import BlissScan
# from tomwer.core.scan.hdf5scan import HDF5TomoScan
# from pypushflow.representation.scheme.ows_parser import OwsParser
# from tomwer.core.process.reconstruction.nabu import settings as nabu_settings
# from silx.io.utils import h5py_read_dataset
# from ..process import _exec
# from glob import glob
# import os
#
#
# class TestNxTomoDarkRef(unittest.TestCase):
#     """
#     Test the execution of the following workflow:
#     nxtomomill -> dark and flat field
#     uses:
#
#     - 'nxtomomill_darkref.ows': where dark is not computed and ref is average
#     - 'nxtomomill_darkref_2.ows': where dark is average and ref is median
#     """
#
#     TIMEOUT = 120
#
#     def setUp(self) -> None:
#         unittest.TestCase.setUp(self)
#         self._data_dir = tempfile.mkdtemp()
#         self._bliss_mock = MockBlissAcquisition(
#             n_sample=1,
#             n_sequence=1,
#             n_darks=10,
#             n_flats=10,
#             n_scan_per_sequence=2,
#             output_dir=self._data_dir,
#         )
#         self.master_file = self._bliss_mock.samples[0].sample_file
#         assert os.path.exists(self.master_file)
#         self.bliss_scan = BlissScan(
#             master_file=self.master_file, entry="/1.1", proposal_file=None
#         )
#         self.tomwer_file = os.path.join(
#             self._bliss_mock.samples[0].path, "tomwer_processes.h5"
#         )
#
#     @staticmethod
#     def get_scheme(file_path):
#         return OwsParser.scheme_load(file_path, load_handlers=True)
#
#     def tearDown(self) -> None:
#         if os.path.isdir(self._data_dir):
#             shutil.rmtree(self._data_dir)
#
#     def testProcessing_1(self):
#         self.assertFalse(os.path.exists(self.tomwer_file))
#         scheme_file = UtilsTest.getOrangeExecTestFile("nxtomomill_darkref.ows")
#         self.scheme = self.get_scheme(scheme_file)
#         _exec(scheme=self.scheme, scan=self.bliss_scan, timeout=self.TIMEOUT)
#         self.assertTrue(os.path.exists(self.tomwer_file))
#         with h5py.File(self.tomwer_file, "r") as h5s:
#             self.assertTrue("entry0000" in h5s)
#             self.assertTrue("entry0000/tomwer_process_0/results/flats" in h5s)
#             self.assertTrue(
#                 h5s["entry0000/tomwer_process_0/results/flats/100"].shape, (2560, 2160)
#             )
#             self.assertFalse("entry0000/tomwer_process_0/results/darks" in h5s)
#
#     def testProcessing_2(self):
#         self.assertFalse(os.path.exists(self.tomwer_file))
#         scheme_file = UtilsTest.getOrangeExecTestFile("nxtomomill_darkref_2.ows")
#         self.scheme = self.get_scheme(scheme_file)
#         _exec(scheme=self.scheme, scan=self.bliss_scan, timeout=self.TIMEOUT)
#         self.assertTrue(os.path.exists(self.tomwer_file))
#         with h5py.File(self.tomwer_file, "r") as h5s:
#             self.assertTrue("entry0000" in h5s)
#             self.assertTrue("entry0000/tomwer_process_0/results/flats" in h5s)
#             self.assertTrue(
#                 h5s["entry0000/tomwer_process_0/results/flats/100"].shape, (2560, 2160)
#             )
#             self.assertTrue("entry0000/tomwer_process_0/results/darks" in h5s)
#             self.assertTrue(
#                 h5s["entry0000/tomwer_process_0/results/darks/0"].shape, (2560, 2160)
#             )
#
#
# class TestDarkRefAxisNabu(unittest.TestCase):
#     """
#     Test the execution of the following workflow:
#     dark and flat field -> center of rotation -> nabu reconstruction
#     uses:
#
#     - 'darkref_axis_nabu.ows'
#         - dark: average, ref: median
#         - cor: manual at 70
#         - nabu: reconstruct slice 20
#     - 'darkref_axis_nabu_2.ows'
#         - dark: average, ref: median
#         - cor: use centered (value should be ~79)
#         - nabu: reconstruct middle slice (1080)
#     """
#
#     TIMEOUT = 120
#
#     def setUp(self) -> None:
#         unittest.TestCase.setUp(self)
#         self._data_dir = tempfile.mkdtemp()
#         self.master_file = os.path.join(self._data_dir, "frm_edftomomill_twoentries.nx")
#         self.tomwer_file = os.path.join(
#             self._data_dir, "frm_edftomomill_twoentries_tomwer_processes.h5"
#         )
#
#         shutil.copyfile(
#             UtilsTest.getH5Dataset(folderID="frm_edftomomill_twoentries.nx"),
#             self.master_file,
#         )
#         self.scan = HDF5TomoScan(scan=self.master_file, entry="entry0000")
#         self._nabu_cfg_folder = os.path.join(
#             self.scan.path, nabu_settings.NABU_CFG_FILE_FOLDER
#         )
#
#     @staticmethod
#     def get_scheme(file_path):
#         return OwsParser.scheme_load(file_path, load_handlers=True)
#
#     def tearDown(self) -> None:
#         if os.path.isdir(self._data_dir):
#             shutil.rmtree(self._data_dir)
#
#     def testProcessing_1(self):
#         self.assertFalse(os.path.exists(self.tomwer_file))
#         self.assertEqual(len(glob(self._nabu_cfg_folder + os.sep + "*.cfg")), 0)
#         scheme_file = UtilsTest.getOrangeExecTestFile("darkref_axis_nabu.ows")
#         self.scheme = self.get_scheme(scheme_file)
#         # monkey patch parameters to use nabu in dry tun mode
#         self.scheme.nodes[2].properties["_rpSetting"]["dry_run"] = True
#         _exec(scheme=self.scheme, scan=self.scan, timeout=self.TIMEOUT)
#         self.assertTrue(os.path.exists(self.tomwer_file))
#         with h5py.File(self.tomwer_file, "r") as h5s:
#             # check darks
#             self.assertTrue("entry0000" in h5s)
#             self.assertTrue("entry0000/tomwer_process_0/results/flats" in h5s)
#             self.assertTrue(
#                 h5s["entry0000/tomwer_process_0/results/flats/1"].shape, (2560, 2160)
#             )
#             self.assertTrue("entry0000/tomwer_process_0/results/darks" in h5s)
#             self.assertTrue(
#                 h5s["entry0000/tomwer_process_0/results/darks/0"].shape, (2560, 2160)
#             )
#             # check axis
#             self.assertEqual(
#                 h5py_read_dataset(h5s["entry0000/tomwer_process_1/program"]),
#                 "tomwer_axis",
#             )
#             self.assertAlmostEqual(
#                 h5py_read_dataset(
#                     h5s["entry0000/tomwer_process_1/results/center_of_rotation"]
#                 ),
#                 70.0,
#                 places=4,
#             )
#
#             # check nabu
#             self.assertEqual(
#                 h5py_read_dataset(h5s["entry0000/tomwer_process_2/program"]),
#                 "nabu-slices",
#             )
#             axp = h5py_read_dataset(
#                 h5s[
#                     "entry0000/tomwer_process_2/configuration/reconstruction/rotation_axis_position"
#                 ]
#             )
#             self.assertAlmostEqual(float(axp), (10.0 + 70.0), places=4)
#             # +10: moved to nabu system (half width
#         # check nabu files
#         self.assertEqual(len(glob(self._nabu_cfg_folder + os.sep + "*.cfg")), 2)
#
#     def testProcessing_2(self):
#         self.assertFalse(os.path.exists(self.tomwer_file))
#         self.assertEqual(len(glob(self._nabu_cfg_folder + os.sep + "*.cfg")), 0)
#
#         scheme_file = UtilsTest.getOrangeExecTestFile("darkref_axis_nabu_2.ows")
#         self.scheme = self.get_scheme(scheme_file)
#         # monkey patch parameters to use nabu in dry tun mode
#         self.scheme.nodes[2].properties["_rpSetting"]["dry_run"] = True
#
#         _exec(scheme=self.scheme, scan=self.scan, timeout=self.TIMEOUT)
#         self.assertTrue(os.path.exists(self.tomwer_file))
#         with h5py.File(self.tomwer_file, "r") as h5s:
#             # check darks
#             self.assertTrue("entry0000" in h5s)
#             self.assertTrue("entry0000/tomwer_process_0/results/flats" in h5s)
#             self.assertTrue(
#                 h5s["entry0000/tomwer_process_0/results/flats/1"].shape, (2560, 2160)
#             )
#             self.assertTrue("entry0000/tomwer_process_0/results/darks" in h5s)
#             self.assertTrue(
#                 h5s["entry0000/tomwer_process_0/results/darks/0"].shape, (2560, 2160)
#             )
#             # check axis
#             self.assertEqual(
#                 h5py_read_dataset(h5s["entry0000/tomwer_process_1/program"]),
#                 "tomwer_axis",
#             )
#             axis_cor = h5py_read_dataset(
#                 h5s["entry0000/tomwer_process_1/results/center_of_rotation"]
#             )
#
#             # check nabu
#             self.assertEqual(
#                 h5py_read_dataset(h5s["entry0000/tomwer_process_2/program"]),
#                 "nabu-slices",
#             )
#             axp = h5py_read_dataset(
#                 h5s[
#                     "entry0000/tomwer_process_2/configuration/reconstruction/rotation_axis_position"
#                 ]
#             )
#             self.assertAlmostEqual(float(axp), (10.0 + axis_cor), places=4)
#             # +10: moved to nabu system (half width
#         # check nabu files
#         self.assertEqual(len(glob(self._nabu_cfg_folder + os.sep + "*.cfg")), 2)
#
#
# class TestPythonScript(unittest.TestCase):
#     """
#     Test the execution of the following workflow:
#     Data list -> Python script test -> Python script test 2
#
#     Data list will be ignored
#     The first scripts create a 'test.txt' file at the root level
#     The second scripts create a 'test2.txt' file at the root level
#
#     Make sure the code is correctly executed and that it.
#     """
#
#     TIMEOUT = 120
#
#     def setUp(self) -> None:
#         unittest.TestCase.setUp(self)
#         self._data_dir = tempfile.mkdtemp()
#         self.master_file = os.path.join(self._data_dir, "frm_edftomomill_twoentries.nx")
#         self.tomwer_file = os.path.join(self._data_dir, "tomwer_processes.h5")
#
#         shutil.copyfile(
#             UtilsTest.getH5Dataset(folderID="frm_edftomomill_twoentries.nx"),
#             self.master_file,
#         )
#         self.scan = HDF5TomoScan(scan=self.master_file, entry="entry0000")
#
#     @staticmethod
#     def get_scheme(file_path):
#         return OwsParser.scheme_load(file_path, load_handlers=True)
#
#     def tearDown(self) -> None:
#         if os.path.isdir(self._data_dir):
#             shutil.rmtree(self._data_dir)
#
#     def testProcessing(self):
#         self.assertEqual(len(glob(self.scan.path + os.sep + "test*.txt")), 0)
#         scheme_file = UtilsTest.getOrangeExecTestFile("test_tomwer_python_scripts.ows")
#         self.scheme = self.get_scheme(scheme_file)
#         _exec(scheme=self.scheme, scan=self.scan, timeout=self.TIMEOUT)
#         self.assertEqual(len(glob(self.scan.path + os.sep + "test*.txt")), 2)
#
#
# def suite():
#     test_suite = unittest.TestSuite()
#     for ui in (TestNxTomoDarkRef, TestDarkRefAxisNabu, TestPythonScript):
#         test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
#     return test_suite
#
#
# if __name__ == "__main__":
#     unittest.main(defaultTest="suite")



# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

"""contain utils for score process
"""

__authors__ = [
    "H.Payno",
]
__license__ = "MIT"
__date__ = "12/05/2021"


from typing import Union
from tomwer.core.scan.edfscan import EDFTomoScan
from ..nabu import settings as nabu_settings
from tomwer.core.process.reconstruction.nabu.nabuslices import (
    _get_file_basename_reconstruction,
    generate_nabu_configfile,
)
import subprocess
from tomwer.io.utils import format_stderr_stdout
from . import utils
import logging
import os

try:
    from nabu.app.local_reconstruction import LocalReconstruction
except ImportError as e:
    has_nabu = False
else:
    has_nabu = True

_logger = logging.getLogger(__name__)


def run_nabu_one_slice_several_config(
    scan,
    nabu_configs,
    dry_run,
    slice_index: Union[int, str],
    local,
    file_format,
    advancement=None,
) -> tuple:
    """
    # TODO: might need something like a context or an option "keep" slice in memory

    :param scan:
    :param nabu_configs: list of nabu configurations to be run
    :param dry_run:
    :param int slice_index: slice index to reconstruct or "middle"
    :param local:
    :param advancement: optional Progress class to display advancement
    :return: tuple with (success, recons_urls (list of output urls),
             tuple of outs, tuples of errs)
    """
    if slice_index == "middle":
        if scan.dim_2 is not None:
            slice_index = scan.dim_2 // 2
        else:
            _logger.warning(
                "scan.dim_2 returns None, unable to deduce middle " "pick 1024"
            )
            slice_index = 1024
    assert isinstance(slice_index, int), "slice_index should be an int"

    def preprocess_config(config, cor: float):
        dataset_params = scan.get_nabu_dataset_info()
        if "dataset" in config:
            dataset_params.update(config["dataset"])
        config["dataset"] = dataset_params

        if local is True:
            resources_method = "local"
        else:
            resources_method = "slurm"
        config["resources"] = utils.get_nabu_resources_desc(
            scan=scan, workers=1, method=resources_method
        )
        # force overwrite results
        if "output" not in config:
            config["output"] = {}
        config["output"].update({"overwrite_results": 1})

        def treateOutputConfig(_config):
            """
            - add or overwrite some parameters of the dictionary
            - create the output directory if does not exist
            """
            pag = False
            db = None
            if "phase" in _config:
                if "method" in _config["phase"] and _config["phase"]["method"] != "":
                    pag = True
                    if "delta_beta" in _config["phase"]:
                        db = round(float(_config["phase"]["delta_beta"]))
            if "output" in config:
                _file_name = _get_file_basename_reconstruction(
                    scan=scan, slice_index=slice_index, pag=pag, db=db
                )
                _config["output"]["file_prefix"] = "cor_{}_{}".format(_file_name, cor)
                if _config["output"]["location"] not in ("", None):
                    # if user specify the location
                    if not os.path.isdir(_config["output"]["location"]):
                        os.makedirs(_config["output"]["location"])
                else:
                    # otherwise default location will be the data root level
                    _config["output"]["location"] = os.sep.join(
                        scan.path, "saaxis_results"
                    )
            if "reconstruction" not in _config:
                _config["reconstruction"] = {}
            _config["reconstruction"]["start_z"] = slice_index
            _config["reconstruction"]["end_z"] = slice_index
            return _config

        config = treateOutputConfig(config)
        # the policy is to save nabu .cfg file at the same location as the
        # force overwrite results

        cfg_folder = os.path.join(
            config["output"]["location"],
            nabu_settings.NABU_CFG_FILE_FOLDER,
        )
        if not os.path.exists(cfg_folder):
            os.makedirs(cfg_folder)

        name = (
            config["output"]["file_prefix"] + nabu_settings.NABU_CONFIG_FILE_EXTENSION
        )
        if not isinstance(scan, EDFTomoScan):
            name = "_".join((scan.entry, name))
        conf_file = os.path.join(cfg_folder, "cor_{}_{}".format(cor, name))
        return config, conf_file

    recons_urls = {}
    outs = []
    errs = []
    if advancement:
        advancement.setMaxAdvancement(len(nabu_configs))
    # TODO: add logger and extra options
    for cor, config in nabu_configs.items():
        config, conf_file = preprocess_config(config, cor)
        # add some tomwer metadata and save the configuration
        # note: for now the section is ignored by nabu but shouldn't stay that way
        with utils.TomwerInfo(config) as config_to_dump:
            generate_nabu_configfile(
                fname=conf_file, config=config_to_dump, options_level="advanced"
            )
        if dry_run:
            continue

        if slice_index is not None and dry_run is False and local:
            if not has_nabu:
                raise ImportError("Fail to import nabu")
            _logger.info(
                "run nabu slice reconstruction for %s with %s" "" % (scan.path, config)
            )

            # need to be executed in his own context
            command = " ".join(
                ("python", "-m", "nabu.resources.cli.reconstruct", conf_file)
            )
            _logger.info('call nabu from "{}"'.format(command))

            process = subprocess.Popen(
                command,
                shell=True,
                cwd=scan.path,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            out, err = process.communicate()
            out_str = format_stderr_stdout(out, err)
            outs.append(out)
            errs.append(err)

        urls = utils.get_recons_urls(
            file_prefix=config_to_dump["output"]["file_prefix"],
            location=config_to_dump["output"]["location"],
            slice_index=None,
            scan=scan,
            file_format=file_format,
            start_z=None,
            end_z=None,
        )
        # specific treatment for cor: rename output files
        recons_urls[cor] = urls
        if advancement:
            advancement.increaseAdvancement(1)
    return (
        len(recons_urls) > 0,
        recons_urls,
        tuple(outs),
        tuple(errs),
    )



# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

"""contain the SAAxisProcess. Half automatic center of rotation calculation
"""

__authors__ = [
    "H.Payno",
]
__license__ = "MIT"
__date__ = "10/02/2021"


from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.process.reconstruction.nabu.nabuslices import (
    interpret_tomwer_configuration,
)
from typing import Iterable
from ..nabu import utils
from tomwer.core.process.reconstruction.scores.params import ScoreMethod
from .params import ReconstructionMode
from .params import SAAxisParams
from tomwer.core.process.baseprocess import SingleProcess, _input_desc, _output_desc
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils import logconfig
from tomwer.core.process.reconstruction.nabu.utils import check_sinogram_half
from processview.core.manager import ProcessManager, DatasetState
from processview.core.superviseprocess import SuperviseProcess
from tomwer.core.process.baseprocess import BaseProcess
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomoscan.io import HDF5File
from silx.io.utils import get_data
import tomwer.version
import numpy
from tomwer.core.progress import Progress
from tomwer.core.process.reconstruction.axis import AxisRP
from typing import Union
import os
import logging
import copy
import h5py
from tomwer.core.process.reconstruction.scores import compute_score
from tomwer.core.process.reconstruction.scores import ComputedScore
from tomwer.core.process.reconstruction.scores import get_disk_mask_radius, apply_roi
from tomwer.core.process.reconstruction.nabu.nabuscores import (
    run_nabu_one_slice_several_config,
)

try:
    from nabu.app.local_reconstruction import LocalReconstruction
except ImportError as e:
    has_nabu = False
else:
    has_nabu = True


_logger = logging.getLogger(__name__)


DEFAULT_RECONS_FOLDER = "saaxis_results"


def one_slice_several_cor(scan, configuration: Union[dict, SAAxisParams]) -> tuple:
    """
    Run a slice reconstruction using nabu per Center Of Rotation (cor) provided
    Then for each compute a score (quality) of the center of rotation

    :param TomwerScanBase scan:
    :param dict configuration:
    :return: cor_reconstructions, outs, errs
             cor_reconstructions is a dictionary of cor as key and a tuple
             (url, score) as value
    :rtype: tuple
    """
    if isinstance(configuration, dict):
        configuration = SAAxisParams.from_dict(configuration)
    elif not isinstance(configuration, SAAxisParams):
        raise TypeError(
            "configuration should be a dictionary or an instance of SAAxisParams"
        )
    check_sinogram_half(scan)

    configuration.check_configuration()
    mode = ReconstructionMode.from_value(configuration.mode)
    slice_index = configuration.slice_indexes
    cors = configuration.cors
    nabu_config = configuration.nabu_params
    output_dir = configuration.output_dir
    dry_run = configuration.dry_run
    nabu_output_config = nabu_config.get("output", {})
    file_format = nabu_output_config.get("file_format", "hdf5")
    _logger.info(
        "launch reconstruction of slice {} and cors {}".format(slice_index, cors)
    )
    if mode is ReconstructionMode.VERTICAL:
        if isinstance(slice_index, str):
            if not slice_index == "middle":
                raise ValueError("slice index {} not recognized".format(slice_index))
        elif not len(slice_index) == 1:
            raise ValueError("{} mode only manage one slice".format(mode.value))
        else:
            slice_index = list(slice_index.values())[0]
        advancement = Progress("saaxis - slice {} of {}".format(slice_index, scan))

        _, cor_reconstructions, outs, errs = run_slice_reconstruction(
            scan=scan,
            slice_index=slice_index,
            cor_positions=cors,
            config=nabu_config,
            output_dir=output_dir,
            dry_run=dry_run,
            file_format=file_format,
            advancement=advancement,
        )
    else:
        raise ValueError("{} is not handled for now".format(mode))

    def load_datasets():
        datasets_ = {}
        for cor, urls in cor_reconstructions.items():
            for url in urls:
                try:
                    data = get_data(url=url)
                except Exception as e:
                    _logger.error(
                        "Fail to compute a score for {}. Reason is {}"
                        "".format(url.path(), str(e))
                    )
                    datasets_[cor] = (url, None)
                else:
                    if data.ndim == 3:
                        if data.shape[0] == 1:
                            data = data.reshape(data.shape[1], data.shape[2])
                        elif data.shape[2] == 1:
                            data = data.reshape(data.shape[0], data.shape[1])
                        else:
                            raise ValueError(
                                "Data is expected to be 2D. Not {}".format(data.ndim)
                            )
                    elif data.ndim == 2:
                        pass
                    else:
                        raise ValueError(
                            "Data is expected to be 2D. Not {}".format(data.ndim)
                        )

                    datasets_[cor] = (url, data)
        return datasets_

    datasets = load_datasets()

    mask_disk_radius = get_disk_mask_radius(datasets)
    scores = {}
    rois = {}
    for cor, (url, data) in datasets.items():
        if data is None:
            score = None
        else:
            assert data.ndim == 2
            data_roi = apply_roi(data=data, radius=mask_disk_radius, url=url)
            rois[cor] = data_roi

            # move data_roi to [0-1] range
            #  preprocessing: get percentile 0 and 99 from image and
            #  "clean" highest and lowest pixels from it
            min_p, max_p = numpy.percentile(data_roi, (1, 99))
            data_roi_int = data_roi[...]
            data_roi_int[data_roi_int < min_p] = min_p
            data_roi_int[data_roi_int > max_p] = max_p
            data_roi_int = (data_roi_int - min_p) / (max_p - min_p)

            score = ComputedScore(
                tv=compute_score(data=data_roi_int, method=ScoreMethod.TV),
                std=compute_score(data=data_roi_int, method=ScoreMethod.STD),
            )
        scores[cor] = (url, score)

    return scores, outs, errs, rois


def run_slice_reconstruction(
    scan: TomwerScanBase,
    cor_positions: Iterable,
    slice_index: int,
    config: dict,
    output_dir=None,
    dry_run: bool = False,
    local: bool = True,
    file_format: str = "hdf5",
    advancement=None,
) -> tuple:
    """
    call nabu for a reconstruction on scan with the given configuration

    :param TomwerScanBase scan: scan to reconstruct
    :param tuple: cor_positions cor position to used for reconstruction
    :param dict config: configuration to run the reconstruction
    :param Union[None,str]: output dir folder. If None then this will be store
                            under the acquisition folder/saaxis_results
    :param bool dry_run: do we want to run dry
    :param bool local: do we want to run a local reconstruction
    :param advancement: optional Progress class to display advancement

    :return: success, recons_urls, outs, errs
             recons_urls is a dict with cor value as key (float) and reconstructed slice url
             as value
    :rtype: dict
    """
    nabu_configurations = interpret_tomwer_configuration(config, scan=None)
    if len(nabu_configurations) == 0:
        raise RuntimeWarning(
            "Unable to get a valid nabu configuration for " "reconstruction."
        )
    elif len(nabu_configurations) > 1:
        _logger.warning(
            "Several configuration found for nabu (you probably "
            "ask for several delta/beta value or several slices). "
            "Picking the first one."
        )
    if output_dir is None:
        output_dir = os.path.join(scan.path, DEFAULT_RECONS_FOLDER)
    if scan.process_file is not None:
        steps_file_basename, _ = os.path.splitext(scan.process_file)
        steps_file_basename = "_".join(
            ("steps_file_basename", "nabu", "sinogram", "save", "step")
        )
        steps_file_basename = steps_file_basename + ".hdf5"
        steps_file = os.path.join(output_dir, steps_file_basename)
    else:
        steps_file = ""

    configs = {}

    for i_cor, cor in enumerate(cor_positions):
        nabu_configuration = copy.deepcopy(nabu_configurations[0][0])
        nabu_configuration["pipeline"] = {
            "save_steps": "sinogram" if i_cor == 0 else "",
            "resume_from_step": "sinogram",
            "steps_file": steps_file,
        }
        # convert cor from tomwer ref to nabu ref
        if scan.dim_1 is not None:
            cor_nabu_ref = cor + scan.dim_1 / 2.0
        else:
            _logger.warning("enable to get image half width. Set it to 1024")
            cor_nabu_ref = cor + 1024
        # handle reconstruction section
        if "reconstruction" not in nabu_configuration:
            nabu_configuration["reconstruction"] = {}
        nabu_configuration["reconstruction"]["rotation_axis_position"] = str(
            cor_nabu_ref
        )
        # handle output section
        if "output" not in nabu_configuration:
            nabu_configuration["output"] = {}
        nabu_configuration["output"]["location"] = output_dir
        nabu_configuration["output"]["file_format"] = file_format
        # handle resources section
        if local is True:
            resources_method = "local"
        else:
            resources_method = "slurm"
        nabu_configuration["resources"] = utils.get_nabu_resources_desc(
            scan=scan, workers=1, method=resources_method
        )
        configs[cor] = nabu_configuration
    return run_nabu_one_slice_several_config(
        nabu_configs=configs,
        scan=scan,
        local=local,
        slice_index=slice_index,
        dry_run=dry_run,
        file_format=file_format,
        advancement=advancement,
    )


class SAAxisProcess(SingleProcess, SuperviseProcess):
    """
    Main process to launch several reconstruction of a single slice with
    several Center Of Rotation (cor) values
    """

    inputs = [
        _input_desc(
            name="data", type=TomwerScanBase, handler="pathReceived", doc="scan path"
        ),
    ]
    outputs = [_output_desc(name="data", type=TomwerScanBase, doc="scan path")]

    def __init__(self, process_id=None, dump_process=True):
        SingleProcess.__init__(self)
        SuperviseProcess.__init__(self, process_id=process_id)
        self._dry_run = False
        self._dump_process = dump_process
        self._std_outs = tuple()
        self._std_errs = tuple()
        self._dump_roi = False

    @property
    def std_outs(self):
        return self._std_outs

    @property
    def std_errs(self):
        return self._std_errs

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    @property
    def dry_run(self):
        return self._dry_run

    @property
    def dump_roi(self):
        return self._dump_roi

    @dump_roi.setter
    def dump_roi(self, dump):
        self._dump_roi = dump

    def set_configuration(self, configuration: dict) -> None:
        return self.set_properties(configuration)

    def set_properties(self, properties):
        if isinstance(properties, SAAxisParams):
            self._settings = properties.to_dict()
        elif isinstance(properties, dict):
            self._settings = properties
        else:
            raise TypeError(
                "properties should be an instance of dict or " "SAAxisParams"
            )

    @staticmethod
    def autofocus(scan):
        scores = scan.saaxis_params.scores
        score_method = scan.saaxis_params.score_method
        best_cor, best_score = None, 0
        for cor, (_, score_cls) in scores.items():
            score = score_cls.get(score_method)
            if score > best_score:
                best_cor, best_score = cor, score
        scan.saaxis_params.autofocus = best_cor
        if scan.axis_params is None:
            scan.axis_params = AxisRP()
        scan.axis_params.frame_width = scan.dim_1
        scan.axis_params.set_relative_value(best_cor)

    def process(self, scan=None):
        if scan is None:
            return None
        if isinstance(scan, TomwerScanBase):
            scan = scan
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise ValueError(
                "input type of {}: {} is not managed" "".format(scan, type(scan))
            )
        # TODO: look and update if there is some nabu reconstruction
        # or axis information to be used back
        configuration = self.get_configuration()
        params = SAAxisParams.from_dict(configuration)
        # insure output dir is created
        if params.output_dir in (None, ""):
            params.output_dir = os.path.join(scan.path, "saaxis_results")
            if not os.path.exists(params.output_dir):
                os.makedirs(params.output_dir)
        # try to find an estimated cor
        #  from a previously computed cor
        if params.estimated_cor is None and scan.axis_params is not None:
            relative_cor = scan.axis_params.relative_cor_value
            if relative_cor is not None and numpy.issubdtype(
                type(relative_cor), numpy.number
            ):
                params.estimated_cor = relative_cor
                _logger.info(
                    "{}: set estimated cor from previously computed cor ({})".format(
                        str(scan), params.estimated_cor
                    )
                )
        #  from scan.estimated_cor_position
        if params.estimated_cor is None and scan.estimated_cor_frm_motor is not None:
            params.estimated_cor = scan.estimated_cor_frm_motor
            _logger.info(
                "{}: set estimated cor from motor position ({})".format(
                    str(scan), params.estimated_cor
                )
            )
        if scan.dim_1 is not None:
            params.image_width = scan.dim_1
        scan.saaxis_params = params
        cors_res, self._std_outs, self._std_errs, rois = one_slice_several_cor(
            scan=scan,
            configuration=self.get_configuration(),
        )
        scan.saaxis_params.scores = cors_res
        self.autofocus(scan=scan)
        self._process_end(scan=scan, cors_res=cors_res, score_rois=rois)
        return scan

    def _process_end(self, scan, cors_res, score_rois):
        assert isinstance(scan, TomwerScanBase)
        try:
            extra = {
                logconfig.DOC_TITLE: self._scheme_title,
                logconfig.SCAN_ID: str(scan),
            }
            slice_index = self.get_configuration().get("slice_index", None)

            if cors_res is None:
                info = "fail to compute cor scores of slice {} for scan {}." "".format(
                    slice_index, scan
                )
                _logger.processFailed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan, process=self, state=DatasetState.FAILED, details=info
                )
            else:
                info = "cor scores of slice {} for scan {} computed." "".format(
                    slice_index, scan
                )
                _logger.processSucceed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan,
                    process=self,
                    state=DatasetState.WAIT_USER_VALIDATION,
                    details=info,
                )
        except Exception as e:
            _logger.error(e)
        else:
            if self._dump_process:
                process_idx = SAAxisProcess.process_to_tomwer_processes(
                    scan=scan,
                )
                if self.dump_roi and process_idx is not None:
                    self.dump_rois(
                        scan, score_rois=score_rois, process_index=process_idx
                    )

    @staticmethod
    def dump_rois(scan, score_rois, process_index):
        process_file = scan.process_file
        process_name = "tomwer_process_" + str(process_index)

        if scan.saaxis_params.scores in (None, {}):
            return

        def get_process_path():
            return "/".join((scan.entry or "entry", process_name))

        # save it to the file
        with BaseProcess._get_lock(process_file):
            # needs an extra lock for multiprocessing

            with HDF5File(process_file, mode="a") as h5f:
                nx_process = h5f.require_group(get_process_path())
                score_roi_grp = nx_process.require_group("score_roi")
                for cor, roi in score_rois.items():
                    score_roi_grp[str(cor)] = roi
                    score_roi_grp[str(cor)].attrs["interpretation"] = "image"

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        return "semi-automatic axis"

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return tomwer.version.version

    @staticmethod
    def definition():
        """definition of the process"""
        return "Semi automatic center of rotation / axis calculation"

    @staticmethod
    def process_to_tomwer_processes(scan):
        if scan.process_file is not None:
            entry = "entry"
            if isinstance(scan, HDF5TomoScan):
                entry = scan.entry

            cor = None
            if hasattr(scan, "axis_params"):
                cor = scan.axis_params.relative_cor_value

            process_index = scan.pop_process_index()
            with scan.acquire_process_file_lock():
                BaseProcess._register_process(
                    process_file=scan.process_file,
                    entry=entry,
                    results={"center_of_rotation": cor if cor is not None else "-"},
                    configuration=scan.saaxis_params.to_dict(),
                    process_index=process_index,
                    overwrite=True,
                    process=SAAxisProcess,
                )
                SAAxisProcess._extends_results(
                    scan=scan, entry=entry, process_index=process_index
                )
            return process_index

    @staticmethod
    def _extends_results(scan, entry, process_index):
        process_file = scan.process_file
        process_name = "tomwer_process_" + str(process_index)

        if scan.saaxis_params.scores in (None, {}):
            return

        def get_process_path():
            return "/".join((entry or "entry", process_name))

        # save it to the file
        with BaseProcess._get_lock(process_file):
            # needs an extra lock for multiprocessing

            with HDF5File(process_file, mode="a") as h5f:
                nx_process = h5f.require_group(get_process_path())
                if "NX_class" not in nx_process.attrs:
                    nx_process.attrs["NX_class"] = "NXprocess"

                results = nx_process.require_group("results")
                for cor, (url, score) in scan.saaxis_params.scores.items():
                    results_cor = results.require_group(str(cor))
                    for method in ScoreMethod:
                        results_cor[method.value] = score.get(method)

                    link_path = os.path.relpath(
                        url.file_path(),
                        os.path.dirname(process_file),
                    )
                    results_cor["reconstructed_slice"] = h5py.ExternalLink(
                        link_path, url.data_path()
                    )



# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

"""contain the SADeltaBetaProcess. Half automatic best delta / beta finder
"""

__authors__ = [
    "H.Payno",
]
__license__ = "MIT"
__date__ = "12/05/2021"


from tomwer.core.scan.scanfactory import ScanFactory
from .params import SADeltaBetaParams
from tomwer.core.process.baseprocess import SingleProcess, _input_desc, _output_desc
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils import logconfig
from tomwer.core.process.reconstruction.nabu.utils import check_sinogram_half
from processview.core.manager import ProcessManager, DatasetState
from processview.core.superviseprocess import SuperviseProcess
from tomwer.core.process.baseprocess import BaseProcess
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomoscan.io import HDF5File
from silx.io.utils import get_data
import tomwer.version
from tomwer.core.process.reconstruction.nabu.nabuslices import (
    run_single_slice_reconstruction,
)
from tomwer.core.progress import Progress
from typing import Union
import os
import logging
import h5py
import numpy
from tomwer.core.process.reconstruction.scores import compute_score
from tomwer.core.process.reconstruction.scores import ScoreMethod
from tomwer.core.process.reconstruction.scores import ComputedScore
from tomwer.core.process.reconstruction.scores import get_disk_mask_radius, apply_roi

try:
    from nabu.app.local_reconstruction import LocalReconstruction
except ImportError as e:
    has_nabu = False
else:
    has_nabu = True


_logger = logging.getLogger(__name__)


DEFAULT_RECONS_FOLDER = "sadeltabeta_results"


def one_slice_several_db(scan, configuration: Union[dict, SADeltaBetaParams]) -> tuple:
    """
    Run a slice reconstruction using nabu per Center Of Rotation (cor) provided
    Then for each compute a score (quality) of the center of rotation

    :param TomwerScanBase scan:
    :param dict configuration:
    :return: cor_reconstructions, outs, errs
             cor_reconstructions is a dictionary of cor as key and a tuple
             (url, score) as value
    :rtype: tuple
    """
    if isinstance(configuration, dict):
        configuration = SADeltaBetaParams.from_dict(configuration)
    elif not isinstance(configuration, SADeltaBetaParams):
        raise TypeError(
            "configuration should be a dictionary or an instance of SAAxisParams"
        )
    check_sinogram_half(scan)

    if scan.axis_params is None:
        from tomwer.core.process.reconstruction.axis import AxisRP

        scan.axis_params = AxisRP()

    configuration.check_configuration()
    slice_index = configuration.slice_indexes
    delta_beta_s = configuration.delta_beta_values
    output_dir = configuration.output_dir
    dry_run = configuration.dry_run
    _logger.info(
        "launch reconstruction of slice {} and delta / beta: {}".format(
            slice_index, delta_beta_s
        )
    )
    if isinstance(slice_index, str):
        if not slice_index == "middle":
            raise ValueError("slice index {} not recognized".format(slice_index))
    elif not len(slice_index) == 1:
        raise ValueError("only manage one slice")
    else:
        slice_index = list(slice_index.values())[0]
    advancement = Progress("sa-delta-beta - slice {} of {}".format(slice_index, scan))

    config = configuration.nabu_params.copy()

    check_sinogram_half(scan)
    _logger.info("start reconstruction of {}".format(str(scan)))
    # if scan contains some center of position copy it to nabu
    if scan.axis_params is not None and scan.axis_params.value_ref_tomwer is not None:
        if "reconstruction" in config:
            # move the cor value to the nabu reference
            cor_nabu_ref = scan.axis_params.value_ref_tomwer + scan.dim_1 / 2.0
            config["reconstruction"]["rotation_axis_position"] = str(cor_nabu_ref)

    _logger.info("set nabu reconstruction parameters to {}".format(str(scan)))
    scan.nabu_recons_params = config

    db_reconstructions = {}
    stderrs = []
    stdouts = []
    all_succeed = True
    if advancement is not None:
        advancement.setMaxAdvancement(len(delta_beta_s))
    for db in delta_beta_s:
        if "output" not in config:
            config["output"] = {}
        if output_dir is None:
            config["output"]["location"] = os.path.join(
                scan.path, DEFAULT_RECONS_FOLDER
            )
        else:
            config["output"]["location"] = output_dir
        # TODO: allow file format modifications
        config["output"]["file_format"] = "hdf5"
        if "phase" not in config:
            config["phase"] = {}
        config["phase"]["delta_beta"] = db
        config["phase"]["method"] = "Paganin"
        succeed, urls, stdout, stderr, f_config = run_single_slice_reconstruction(
            config=config,
            scan=scan,
            local=True,
            slice_index=slice_index,
            dry_run=dry_run,
            ask_sinogram_registration=False,
            ask_sinogram_load=False,
        )
        if len(urls) > 0:
            assert len(urls) == 1, "only one slice expected"
            db_reconstructions[db] = urls[0]
            # if slice_index is None this mean that we are simply creating the
            # .cfg file for nabu full volume.
            if slice_index is not None:
                stderrs.append(stderr)
                stdouts.append(stdout)
            all_succeed = all_succeed and succeed
            if advancement is not None:
                advancement.increaseAdvancement(1)

    def load_datasets():
        datasets_ = {}
        for db, url in db_reconstructions.items():
            try:
                data = get_data(url=url)
            except Exception as e:
                _logger.error(
                    "Fail to compute a score for {}. Reason is {}"
                    "".format(url.path(), str(e))
                )
                datasets_[db] = (url, None)
            else:
                if data.ndim == 3:
                    if data.shape[0] == 1:
                        data = data.reshape(data.shape[1], data.shape[2])
                    elif data.shape[2] == 1:
                        data = data.reshape(data.shape[0], data.shape[1])
                    else:
                        raise ValueError(
                            "Data is expected to be 2D. Not {}".format(data.ndim)
                        )
                elif data.ndim == 2:
                    pass
                else:
                    raise ValueError(
                        "Data is expected to be 2D. Not {}".format(data.ndim)
                    )

                datasets_[db] = (url, data)
        return datasets_

    datasets = load_datasets()

    mask_disk_radius = get_disk_mask_radius(datasets)
    scores = {}
    rois = {}
    for cor, (url, data) in datasets.items():
        if data is None:
            score = None
        else:
            assert data.ndim == 2
            data_roi = apply_roi(data=data, radius=mask_disk_radius, url=url)
            rois[cor] = data_roi

            # move data_roi to [0-1] range
            #  preprocessing: get percentile 0 and 99 from image and
            #  "clean" highest and lowest pixels from it
            min_p, max_p = numpy.percentile(data_roi, (1, 99))
            data_roi_int = data_roi[...]
            data_roi_int[data_roi_int < min_p] = min_p
            data_roi_int[data_roi_int > max_p] = max_p
            data_roi_int = (data_roi_int - min_p) / (max_p - min_p)

            score = ComputedScore(
                tv=compute_score(data=data_roi_int, method=ScoreMethod.TV),
                std=compute_score(data=data_roi_int, method=ScoreMethod.STD),
            )
        scores[cor] = (url, score)

    return scores, stdouts, stderrs, rois


class SADeltaBetaProcess(SingleProcess, SuperviseProcess):
    """
    Main process to launch several reconstruction of a single slice with
    several Center Of Rotation (cor) values
    """

    inputs = [
        _input_desc(
            name="data", type=TomwerScanBase, handler="pathReceived", doc="scan path"
        ),
    ]
    outputs = [_output_desc(name="data", type=TomwerScanBase, doc="scan path")]

    def __init__(self, process_id=None, dump_process=True):
        SingleProcess.__init__(self)
        SuperviseProcess.__init__(self, process_id=process_id)
        self._dry_run = False
        self._dump_process = dump_process
        self._std_outs = tuple()
        self._std_errs = tuple()
        self._dump_roi = False

    @property
    def dump_roi(self):
        return self._dump_roi

    @dump_roi.setter
    def dump_roi(self, dump):
        self._dump_roi = dump

    @property
    def std_outs(self):
        return self._std_outs

    @property
    def std_errs(self):
        return self._std_errs

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    @property
    def dry_run(self):
        return self._dry_run

    def set_configuration(self, configuration: dict) -> None:
        return self.set_properties(configuration)

    def set_properties(self, properties):
        if isinstance(properties, SADeltaBetaParams):
            self._settings = properties.to_dict()
        elif isinstance(properties, dict):
            self._settings = properties
        else:
            raise TypeError(
                "properties should be an instance of dict or " "SAAxisParams"
            )

    @staticmethod
    def autofocus(scan):
        scores = scan.sa_delta_beta_params.scores
        score_method = scan.sa_delta_beta_params.score_method
        best_db, best_score = None, 0
        for cor, (_, score_cls) in scores.items():
            score = score_cls.get(score_method)
            if score > best_score:
                best_db, best_score = cor, score
        scan.sa_delta_beta_params.autofocus = best_db
        scan.sa_delta_beta_params.value = best_db

    def process(self, scan=None):
        if scan is None:
            return None
        if isinstance(scan, TomwerScanBase):
            scan = scan
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise ValueError(
                "input type of {}: {} is not managed" "".format(scan, type(scan))
            )
        # insure scan contains some parameter regarding sa delta / beta
        if scan.sa_delta_beta_params is None:
            scan.sa_delta_beta_params = SADeltaBetaParams()
        # TODO: look and update if there is some nabu reconstruction
        # or axis information to be used back
        configuration = self.get_configuration()
        params = SADeltaBetaParams.from_dict(configuration)
        # insure output dir is created
        if params.output_dir in (None, ""):
            params.output_dir = os.path.join(scan.path, DEFAULT_RECONS_FOLDER)
            if not os.path.exists(params.output_dir):
                os.makedirs(params.output_dir)
        db_res, self._std_outs, self._std_errs, rois = one_slice_several_db(
            scan=scan,
            configuration=params,
        )
        scan.sa_delta_beta_params.scores = db_res
        self.autofocus(scan=scan)
        self._process_end(scan=scan, db_res=db_res, score_rois=rois)
        return scan

    def _process_end(self, scan, db_res, score_rois):
        assert isinstance(scan, TomwerScanBase)
        try:
            extra = {
                logconfig.DOC_TITLE: self._scheme_title,
                logconfig.SCAN_ID: str(scan),
            }
            slice_index = self.get_configuration().get("slice_index", None)

            if db_res is None:
                info = (
                    "fail to compute delta/beta scores of slice {} for scan {}."
                    "".format(slice_index, scan)
                )
                _logger.processFailed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan, process=self, state=DatasetState.FAILED, details=info
                )
            else:
                info = "delta/beta scores of slice {} for scan {} computed." "".format(
                    slice_index, scan
                )
                _logger.processSucceed(info, extra=extra)
                ProcessManager().notify_dataset_state(
                    dataset=scan,
                    process=self,
                    state=DatasetState.WAIT_USER_VALIDATION,
                    details=info,
                )
        except Exception as e:
            _logger.error(e)
        else:
            if self._dump_process:
                process_idx = SADeltaBetaProcess.process_to_tomwer_processes(
                    scan=scan,
                )
                if self.dump_roi and process_idx is not None:
                    self.dump_rois(
                        scan, score_rois=score_rois, process_index=process_idx
                    )

    @staticmethod
    def dump_rois(scan, score_rois: dict, process_index: int):
        if score_rois is None or len(score_rois) == 0:
            return
        if not isinstance(score_rois, dict):
            raise TypeError("score_rois is expected to be a dict")
        process_file = scan.process_file
        process_name = "tomwer_process_" + str(process_index)

        if scan.saaxis_params.scores in (None, {}):
            return

        def get_process_path():
            return "/".join((scan.entry or "entry", process_name))

        # save it to the file
        with BaseProcess._get_lock(process_file):
            # needs an extra lock for multiprocessing

            with HDF5File(process_file, mode="a") as h5f:
                nx_process = h5f.require_group(get_process_path())
                score_roi_grp = nx_process.require_group("score_roi")
                for db, roi in score_rois.items():
                    score_roi_grp[str(db)] = roi
                    score_roi_grp[str(db)].attrs["interpretation"] = "image"

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        return "semi-automatic delta/beta finder"

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return tomwer.version.version

    @staticmethod
    def definition():
        """definition of the process"""
        return "Semi automatic center of rotation / axis calculation"

    @staticmethod
    def process_to_tomwer_processes(scan):
        if scan.process_file is not None:
            entry = "entry"
            if isinstance(scan, HDF5TomoScan):
                entry = scan.entry

            db = None
            if scan.sa_delta_beta_params is not None:
                db = scan.sa_delta_beta_params.selected_delta_beta_value

            process_index = scan.pop_process_index()
            with scan.acquire_process_file_lock():
                BaseProcess._register_process(
                    process_file=scan.process_file,
                    entry=entry,
                    results={"delta_beta": db if db is not None else "-"},
                    configuration=scan.sa_delta_beta_params.to_dict(),
                    process_index=process_index,
                    overwrite=True,
                    process=SADeltaBetaProcess,
                )
                SADeltaBetaProcess._extends_results(
                    scan=scan, entry=entry, process_index=process_index
                )
            return process_index

    @staticmethod
    def _extends_results(scan, entry, process_index):
        process_file = scan.process_file
        process_name = "tomwer_process_" + str(process_index)

        if scan.sa_delta_beta_params.scores in (None, {}):
            return

        def get_process_path():
            return "/".join((entry or "entry", process_name))

        # save it to the file
        with BaseProcess._get_lock(process_file):
            # needs an extra lock for multiprocessing

            with HDF5File(process_file, mode="a") as h5f:
                nx_process = h5f.require_group(get_process_path())
                if "NX_class" not in nx_process.attrs:
                    nx_process.attrs["NX_class"] = "NXprocess"

                results = nx_process.require_group("results")
                for cor, (url, score) in scan.sa_delta_beta_params.scores.items():
                    results_db = results.require_group(str(cor))
                    for method in ScoreMethod:
                        results_db[method.value] = score.get(method)

                    link_path = os.path.relpath(
                        url.file_path(),
                        os.path.dirname(process_file),
                    )
                    results_db["reconstructed_slice"] = h5py.ExternalLink(
                        link_path, url.data_path()
                    )



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/02/2020"

from tomwer.test.utils import skip_gui_test
from tomwer.gui.utils.flow import FlowCanvas
from silx.gui.utils.testutils import TestCaseQt
from silx.gui import qt
import unittest


class ProcessClass:
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name


#
# @pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
# class TestFlowCanvas(TestCaseQt):
#     def setUp(self) -> None:
#         TestCaseQt.setUp(self)
#         self.widget = FlowCanvas(parent=None, direction="vertical")
#
#     def testFlow1(self):
#         self.widget.addProcess(process="preprocessing - step 1", icon=None, draggable=False)
#         self.widget.addProcess(process="preprocessing - step 2", icon=None, draggable=False)
#
#         for name in "process lambda", "process zeta", "process test":
#             self.widget.addProcess(process=name, icon=None)
#
#         style = qt.QApplication.style()
#
#         icon_1 = style.standardIcon(qt.QStyle.SP_DialogApplyButton)
#         icon_2 = style.standardIcon(qt.QStyle.SP_FileLinkIcon)
#
#         for name, icon in zip(("process_ico1", "process_ico2"), (icon_1, icon_2)):
#             self.widget.addProcess(process=name, icon=icon)
#
#         id_process_rm = self.widget.addProcess(process="process to remove")
#         self.widget.removeProcess(id_process_rm)
#
#         for proc_with_class in ("proc_class_1", "proc class b"):
#             self.widget.addProcess(process=ProcessClass(name=proc_with_class))
#
#         self.widget.addProcess(process="postprocessing - step 1", icon=None, draggable=False)
#
#         self.widget.show()
#         # TODO: test the getFlow function result
#
#
# def suite():
#     test_suite = unittest.TestSuite()
#     for ui in (TestFlowCanvas,):
#         test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
#     return test_suite
